
%MACRO calc_cox_iter_stats_dp;

	 /* Wrapper for calculating various Cox regression stats per iteration at a DP site. 
	   Some comments below reference statistical (algorithm) part of the Cox paper using
      Latex like notations.
*/

 %LOCAL freq2 weight2 tbl_events_time_set_dp _tmp_reg_vars_sum _tmp_cox_sscp_q 
              _time_set_vars _new_names_str sscp_all_names;

  %PUT;
   %PUT -----   Start of the calc_cox_iter_stats_dp  iter_nb=&iter_nb. ;
   %PUT;

PROC DATASETS LIB=DATA_IN  NOLIST NOWARN;
   DELETE _REGR_DS_XBETA;
RUN;

 /* Add linear predictor to regression dataset using current estimates of \beta */

PROC SCORE DATA= &reg_ds_dp_in.
					     SCORE=&infolder_lib..&prefix_prev.Model_Coeff	  TYPE=PARMS 
						OUT=_regr_ds_xbeta
		;
   VAR &independent_vars.;
run;

  %LET  freq2=1;
  %LET  weight2=1;

	  %IF  %LENGTH(&freq.) >0  		%THEN %LET  freq2=&freq.;
	  %IF  %LENGTH(&weight.) >0  %THEN 	%LET  weight2=&weight.;

	  /* Define some additional variables in the individul level dataset */
  
	   	   DATA  &prefix_dp.reg_vars;
			  SET _regr_ds_xbeta;
			   
					_RESP_=&dependent_vars.;
				    _FREQ_=&freq2.;
				   _WEIGHT_=&weight2.;

				   dp_cd=&dp_cd.;
			
		    ARRAY indep{*}  &independent_vars. ;

 /*Set _XBETA_  to missing when any of &independent_vars is missing. In most cases, PROC SCORE
	does this correctly. However, when of a regression coefficient is 0 PROC SCORE just ignores it 
	and produces non-missing _XBETA_.  The code below corrects this.*/
	
       DO i=1 TO DIM(indep);
			IF indep{i}=. THEN _XBETA_=.;	
	END;

		IF  _RESP_ = .  OR _XBETA_ = .	OR &censoring_var.= .  THEN  DO; _FREQ_=.; _WEIGHT_=. ; END;

		IF  _FREQ_ >0 AND 	_WEIGHT_ >0;

		_CENSORING_VAR_=(&censoring_var. NE &censoring_lev.);

		         /* Defensive coding. exp(_XBETA_) is a very large number when  _XBETA_ >30 */

			  IF _XBETA_ >30 THEN _XBETA1_ =30;
			  ELSE _XBETA1_=_XBETA_;

		   _FREQ_for_sscp_  =_FREQ_	;
		  _WEIGHT_for_sscp_= _WEIGHT_*exp(_XBETA1_);

		  /* The transformation below is done to avoid warning mrssage from PROC CORR when _FREQ_=1
			  It does not change resulting sscp matrix because it depends on the product of  f*w
		  */
		   IF  _FREQ_=1  THEN
							  DO;
			   _FREQ_for_sscp_=2;
			    _WEIGHT_for_sscp_=0.5*_WEIGHT_for_sscp_;
							END;
		DROP i _XBETA1_;

		RUN;

		
		PROC DATASETS  MEMTYPE=DATA NOLIST;
		   MODIFY &prefix_dp.reg_vars;
		     ATTRIB _ALL_ LABEL=" ";
		     ATTRIB _ALL_ FORMAT=;
		RUN;
		QUIT;
		
 PROC SORT DATA=&prefix_dp.reg_vars;
  BY &strata_vars. &dependent_vars.;
 RUN;

	  %LET _tmp_reg_vars_sum=;

    %IF &iter_nb.=1 %THEN
	                                           %DO;

				 /* Calculate counts of events and sum of weights  per time and strata*/
	  PROC MEANS DATA=&prefix_dp.reg_vars NWAY NOPRINT;
	     BY  &strata_vars. &dependent_vars.;
		   FREQ _FREQ_;
	
			 VAR _CENSORING_VAR_  _WEIGHT_;
	       OUTPUT 	 OUT=	&DPLocal_lib..&prefix0_dp.times_set_cnts( DROP=_TYPE_ )
                            SUM(_CENSORING_VAR_)=_Event_
							SUM(_WEIGHT_)=_WEIGHT_
                               ;
	    RUN;
      QUIT;

	  	 /* Calculate overall counts of records, events and sum of weights  per time and strata*/
	PROC MEANS DATA=&DPLocal_lib..&prefix0_dp.times_set_cnts NWAY NOPRINT;
		 &by_strata_vars. ;
			 VAR _FREQ_ _Event_  _WEIGHT_;
	       OUTPUT 	 OUT=	&msoc_lib..&prefix0.total_cnts( DROP=_TYPE_ _FREQ_)
							SUM(_FREQ_)=Total
                            SUM(_Event_)=_Event_
							SUM(_WEIGHT_)=_WEIGHT_
                               ;
	    RUN;
      QUIT;
	
/* Calculate   weighted sum of events (d_{m,j}), non-weighted count of events (d^(0)_{m,j}) and 
	   sum for each of the covariates (vector d^(1)_{m,j} ) per  strata  and  event time*/ 

   PROC MEANS DATA=&prefix_dp.reg_vars NWAY NOPRINT;
	     BY &strata_vars. &dependent_vars.;
		   FREQ _FREQ_;
           WEIGHT _WEIGHT_;
	
			 VAR &independent_vars.;
			 WHERE _CENSORING_VAR_=1; 
	       OUTPUT 	 OUT=&prefix0_dp.reg_vars_sum1( DROP=_TYPE_   RENAME=(_FREQ_=_Event_ ))
 							SUMWGT=Intercept
                            SUM=
                          ;
	    RUN;
      QUIT;

	  DATA &dplocal_lib..&prefix0_dp.reg_vars_sum1;
	   LENGTH _TYPE_ $ 8 _NAME_ $ 32;
	      SET &prefix0_dp.reg_vars_sum1;
	    
	    _TYPE_="SUM";
        _NAME_= "Intercept" ; 
	  RUN;

	  %LET _tmp_reg_vars_sum=&dplocal_lib..&prefix0_dp.reg_vars_sum1;

					                         %END;	/* Endo do-group for &iter_nb.=1*/

   /*Calculate contributions to matrices S^(l), l=0,1,2 for each strata  and  event time. 

	  The macro  %calculate_sscp uses PROC CORR with SSCP option. The latter implicitly add column of 1 to the input dataset
	  and then calculate all cross-products. In the output dataset the contribution from column of 1 are recorded in the 
	  column "Intercept" and in the row with _NAME_="Intercept"	. The cell with both:
	column "Intercept" and in the row with _NAME_="Intercept"stores contribution to the scalar S^(0), the rest of the column 
	 "Intercept" staores contribution to the scalar S^(1), the rest of the dataset stores contribution to the matrix S^(2).
	  We refer here to contributions to S^(l), l=0,1,2 because we still need to sum over the risk set.

	  Note that column "Intercept" in the output dataset has nothing to do with the intercept in the Cox model. The latter is 
	  remains undefined in the Cox model (it is absorbed into baseline hazard).
	*/

   %calculate_sscp(tbl_name_in=&prefix_dp.reg_vars 
                         ,independent_vars=&independent_vars.
						 ,by_vars=&strata_vars. &dependent_vars.
						 ,freq= _FREQ_for_sscp_ 
                         ,weight=_WEIGHT_for_sscp_
						 ,tbl_name_out=&prefix_dp.cox_sscp);

	 	PROC SORT DATA=&prefix_dp.cox_sscp OUT=&prefix_dp.cox_sscp_srt;
	      BY &strata_vars. &dependent_vars.	;
		   WHERE  _TYPE_ IN ("SSCP") ;
	   RUN;

	/*	If model is stratified by dp_cd then events time set is specific to the DP and is determined here. Otherwise, there is overall time
	     set of events. The latter can be determined only at the AC (central site) which gathers information from all DPs.
	    It can also optionally defined in the call to main macro at the AC*/

	    %IF &pos_dp_cd_in_strata_vars. >0 
         %THEN %LET tbl_events_time_set_dp=&dplocal_lib..&prefix0_dp.all_events_time_set ;
	   %ELSE    %LET tbl_events_time_set_dp=&infolder_lib..&tbl_events_time_set;

	   %isdata(dataset=&tbl_events_time_set_dp.);


		  %IF &nobs.=0  %THEN
		                        %DO;

		PROC SORT DATA=&dplocal_lib..&prefix0_dp.reg_vars_sum1(KEEP=&strata_vars. &dependent_vars.	) 
                              OUT=&dplocal_lib..&prefix0_dp.all_events_time_set NODUPKEY;
		    BY &strata_vars. &dependent_vars.	;
		RUN;

		  %LET tbl_events_time_set_dp=&dplocal_lib..&prefix0_dp.all_events_time_set ;
	                             %END;
		
       %PUT  tbl_events_time_set_dp=&tbl_events_time_set_dp.;
	
	   PROC SORT DATA= &tbl_events_time_set_dp.;
	      BY &strata_vars. &dependent_vars.	;
	   RUN;
	
     /*Merge dataset with all events  times with the dataset that has matrices S^(l), l=0,1,2 
		 and the dataset with both events and censored observations. 
	   */
	DATA	 &prefix_dp.cox_sscp_all_t;
	 	 MERGE &tbl_events_time_set_dp.(IN=InAllEvents)
					   &prefix_dp.cox_sscp_srt(IN=InDpTimesSscp)
					  &DPLocal_lib..&prefix0_dp.times_set_cnts(IN=INDpTimesSet)
					;
			BY &strata_vars.  &dependent_vars.	;

			IN_ALL_EVENTS=InAllEvents;
			 IN_DP_TIMES_sscp=InDpTimesSscp;
             IN_Dp_Times_Set=INDpTimesSet;	

			 ARRAY var_lst  Intercept &independent_vars. ;

	/* At rows with all valid vars names when there is no event at the given DP but there are events at other DPs  */
			 IF InDpTimesSscp=0 THEN
			      DO;
				  _TYPE_="SSCP";
						 DO i=1 TO DIM(var_lst);
						    _NAME_=VNAME(var_lst{i});
					           OUTPUT;
						 END;			
	             END;
			ELSE          OUTPUT;

			 DROP i;
	  RUN;

	  /*
		 PROC FREQ DATA= &prefix_dp.cox_sscp_all_t;
			TABLES IN_ALL_EVENTS*IN_DP_TIMES_sscp* IN_Dp_Times_Set  /MISSING LIST;
		RUN;
	 */

	 %LET _time_set_vars=_FREQ_ _Event_ _WEIGHT_;

	%LET _new_names_str=;

	%modify_colnames(tbl_name=&prefix_dp.cox_sscp_all_t
                                     ,var_lst=Intercept &independent_vars. &_time_set_vars.;
                                    ,prefix=,suffix=__);

	%LET  sscp_all_names=&_new_names_str.;
	%PUT  sscp_all_names=&sscp_all_names.;

	
	 PROC SORT DATA=&prefix_dp.cox_sscp_all_t;
		BY &strata_vars.  _TYPE_  _NAME_  DESCENDING &dependent_vars.	;
	  RUN;

			/* Calculate cumulative sums for S^(l), l=0,1,2 in reverse event time order (over the risk set)*/

	DATA &prefix_dp.cox_sscp_cum_all_e;
		SET	&prefix_dp.cox_sscp_all_t;
		  BY &strata_vars.  _TYPE_  _NAME_   DESCENDING &dependent_vars.	;

		   ARRAY var_lst0 &sscp_all_names.;
		   ARRAY var_lst  Intercept &independent_vars. &_time_set_vars.;

	   RETAIN	 Intercept &independent_vars. &_time_set_vars.;


		  cnt_vars=COUNTW(COMPBL("Intercept &independent_vars."), " ");

	   DO i=1 TO DIM(var_lst);

	   /* Excludes &_time_set_vars. from cumulative sum calculation */

		 IF i <=cnt_vars OR STRIP(lowcase(_NAME_))="intercept" THEN
							  DO;
		  	IF FIRST._NAME_ THEN  var_lst{i}=0;
		  var_lst{i}=SUM(var_lst{i}, var_lst0{i});
							  END;
			 ELSE  var_lst{i}=.;
	   END;

	   IF  IN_ALL_EVENTS THEN OUTPUT;
	   DROP i IN_ALL_EVENTS  IN_DP_TIMES_sscp IN_Dp_Times_Set 
                 &sscp_all_names. cnt_vars;
	 RUN;

   /* For ties=EFRON add matrices Q^(l), l=0,1,2 (see algorithm specs for details) to the final  dataset */

   %LET _tmp_cox_sscp_q=;

   %IF &ties_sh.=e %THEN	 
               %DO;

		 %calculate_sscp(tbl_name_in=&prefix_dp.reg_vars(WHERE=( _CENSORING_VAR_=1)) 
                         ,independent_vars=&independent_vars.
						 ,by_vars=&strata_vars. &dependent_vars.
						 ,freq= _FREQ_for_sscp_ 
                         ,weight=_WEIGHT_for_sscp_
						 ,tbl_name_out=&prefix_dp.cox_sscp_q0);

	DATA  &prefix_dp.cox_sscp_q;
	   SET &prefix_dp.cox_sscp_q0;
	  	 IF _TYPE_ IN ("SSCP") ;

		 _TYPE_="SSCP_Q"   ;

    RUN;

	   %LET _tmp_cox_sscp_q=&prefix_dp.cox_sscp_q;

             %END; 

   %PUT  _tmp_reg_vars_sum=&_tmp_reg_vars_sum.;
   %PUT  _tmp_reg_vars_sum_q=&_tmp_cox_sscp_q.;

   /* Stock together several datasets into one dataset. This simplifies data transfer.
     The contributions from different datasets can be identifiedby the value of the the variable type _TYPE_ . 
      Contrib. from &prefix_dp.cox_sscp_cum_all_e has   _TYPE_="SSCP" and matrices S^(l), l=0,1,2. 
       The macro &_tmp_cox_sscp_q.  is not empty only when Effron approxim. is
     requested. In that case _TYPE_="SSCP_Q".  The macro &_tmp_reg_vars_sum.  is not empty only when
     &iter_nb=1 (it does not depend on beta and tus does not change with iterations). It has  _TYPE_="SUM".
     The dataset has data for all strata and event times.
   */

    DATA 	&dplocal_lib..&prefix_dp.sscp_all_e	;
		 LENGTH dp_cd 8 _TYPE_ $ 8 _NAME_ $ 32;;
	   SET 	&prefix_dp.cox_sscp_cum_all_e
				&_tmp_reg_vars_sum.
				&_tmp_cox_sscp_q.
                  ;
		 dp_cd=COALESCE(dp_cd,&dp_cd.);
	 RUN;

 /*If dp_cd is not part of strata variables then pass &msoc_lib..&prefix.sscp on events times level. 
	Otherwise do summarization over time at each DP and pass to central Gradient and Hessian matrices */
%IF &pos_dp_cd_in_strata_vars. =0 %THEN
	       %DO;

 /* See comment above about the structure of the  dataset &dplocal_lib..&prefix_dp.sscp_all_e.
	Note the dataset &msoc_lib..&prefix.sscp is unrelated to the one used as input to PROC REG to
	solve system of linear equation. The appropriate dataset will be constructed at the central site.
*/

      DATA 	&msoc_lib..&prefix.sscp	;
		 SET &dplocal_lib..&prefix_dp.sscp_all_e;
	 RUN;

			 %END;
%ELSE
			 %DO;

		 /*If dp_cd is  part of strata variables then  calculate DP's contibution to the Hessian and
			gradient. The macro %calc_cox_grad_hessian also calculates some other datasets. 
			 See comments in the macro for details.
			 */


	%calc_cox_grad_hessian(tbl_cox_sscp_in=&dplocal_lib..&prefix_dp.sscp_all_e
											    ,tbl_model_coeff_in =&infolder_lib..&prefix_prev.Model_Coeff
												,tbl_mean_vars_by_st_out=&prefix0_dp.mean_vars_by_st
												,tbl_resid_sum_out=&prefix_dp.resid_sum
												,tbl_baseln_hazard=&prefix0_dp.baseln_hazard
												,tbl_Gradient_out =&prefix_dp.Gradient
												,tbl_Hessian_out =&prefix_dp.Hessian
                                                  ) ;

 /*   Stock together datasets with  Hessian and gradient into one dataset. This simplifies data transfer.
     The contributions from different datasets can be identified 
	  by the value of the the variable type _TYPE_ . 
	  Note the dataset &msoc_lib..&prefix.sscp is unrelated to the one used as input to PROC REG to
	solve system of linear equation. The appropriate dataset will be constructed at the central site.
*/

	     DATA 	&msoc_lib..&prefix.sscp	;
		    LENGTH dp_cd 8 _TYPE_ $ 8 _NAME_ $ 32;;
		    SET  &prefix_dp.Hessian(IN=in_hes)
                     &prefix_dp.Gradient(IN=in_grad)
						;
		   IF            in_hes 	THEN _TYPE_="HESSIAN";
		   ELSE IF 	in_grad    THEN DO; _TYPE_="GRADIENT"; _NAME_="Intercept"; END;

			dp_cd=COALESCE(dp_cd,&dp_cd.);
	    RUN;

		/* The dataset &msoc_lib..&prefix.resid_sum has log-likelihood value for given iteration*/

		 DATA &msoc_lib..&prefix.resid_sum;
		  	LENGTH dp_cd 8;
		     SET &prefix_dp.resid_sum;

		      dp_cd=COALESCE(dp_cd,&dp_cd.);
		RUN;

   %IF   &iter_nb. = 1  %THEN
 								%DO;

   /* The dataset &msoc_lib..&prefix0.mean_vars_by_st  has meand of covariates by strata*/

					   DATA  &msoc_lib..&prefix0.mean_vars_by_st;
						LENGTH dp_cd 8;
						SET &prefix0_dp.mean_vars_by_st;

						   dp_cd=COALESCE(dp_cd,&dp_cd.);
					   RUN;
								 %END;

	 %IF &iter_nb. =%eval(&iter_nb_at_conv. + 1) %THEN
                                %DO;
   /* The dataset &msoc_lib..&prefix0.baseln_hazard  has baseline hazard for each strata
	   and event time.
*/

	               DATA  &msoc_lib..&prefix0.baseln_hazard;
						LENGTH dp_cd 8;
						SET &prefix0_dp.baseln_hazard;
						   dp_cd=COALESCE(dp_cd,&dp_cd.);
						 KEEP dp_cd &strata_vars. &dependent_vars.  _Cum_Hazard0_	;
					   RUN; 
					          %END;

			%END;  /* End of do-group &pos_dp_cd_in_strata_vars.  NE 0*/

%IF &iter_nb. =%eval(&iter_nb_at_conv. + 1) %THEN
                     %DO; 

  /*Save  regression dataset with  final linear predictor values in the DP's dplocal library*/

  DATA &dplocal_lib..&prefix0_dp.reg_vars(SORTEDBY=&strata_vars. &dependent_vars.);
	SET &prefix_dp.reg_vars;
  RUN;
				   %END;

    %PUT;
   %PUT -----   End of the calc_cox_iter_stats_dp  iter_nb=&iter_nb. ;
   %PUT;
   %MEND;