 /*HEADER------------------------------------------------------------------------
	
Calculates HosmerLemeshow for Logistic regression.

Usage:
  %LET prefix0=l1_;
  %calc_hl_stats_central(tbl_name_in=msoc.&prefix0.resid_sum_by_pct
                                            ,tbl_name_hl_partition_out=msoc.&prefix0.hl_partition
                                            ,tbl_name_hl_chisq_out=msoc.&prefix0.HL_ChiSq);

  First version:  01/04/2017, Yury Vilk. 
+------------------------------------------------------------------------HEADER*/


%MACRO calc_hl_stats(tbl_name_in=,tbl_name_hl_partition_out=
                                       ,tbl_name_hl_chisq_out=);

  DATA &tbl_name_hl_partition_out.;
   LENGTH dp_cd Group Total EventsObserved EventsExpected NoneventsObserved NoneventsExpected 
                   _ADJ_VARIANCE__Mean  8;
        SET &tbl_name_in.;
		  	_sml_=1E-12;

		 GROUP=Rank__MEAN_;
		 Total=_FREQ_;
		  EventsObserved=_RESP__Mean*Total;	 
		  NoneventsObserved=Total-EventsObserved;
		  EventsExpected=_MEAN__Mean*Total;
		  NoneventsExpected=Total-EventsExpected;

		    /*Traditionaly HL test uses approx VAR_of_Mean_HL calculated below. 
		     The difference with true VAR_of_Mean is small */
		 _RESIDUAL_SQ=(_RESP__Mean-_MEAN__Mean)**2;
		VAR_of_Mean_HL=MAX(ABS(_MEAN__Mean*(1-_MEAN__Mean)),_sml_)/Total ;
	    VAR_of_Mean=_ADJ_VARIANCE__Mean/Total;

		                         /* HL ChiSq*/
		ChiSq=(_RESP__Mean-_MEAN__Mean)**2/VAR_of_Mean_HL;
		                       /* True ChiSq  */
		ChiSq_t=(_RESP__Mean-_MEAN__Mean)**2/VAR_of_Mean;

		FORMAT EventsExpected NoneventsExpected 8.2;
   	KEEP dp_cd  Group Total EventsObserved EventsExpected NoneventsObserved NoneventsExpected
                    _RESP__Mean _MEAN__Mean _RESIDUAL_SQ VAR_of_Mean_HL  VAR_of_Mean
                     ChiSq ChiSq_t;
  RUN;

  PROC DATASETS NOLIST NOWARN;
	DELETE HL_ChiSq_dp;
  RUN;

  PROC MEANS DATA=&tbl_name_hl_partition_out.  NWAY NOPRINT;
   CLASS dp_cd /missing  ;
   VARS Group Total ChiSq ChiSq_t;
   OUTPUT OUT=HL_ChiSq_dp( DROP=_TYPE_ RENAME=(_FREQ_= N_groups)) 
				 SUM(ChiSq)=ChiSq SUM(ChiSq_t)=ChiSq_t SUM(Total)=Total ;
  RUN;

  DATA &tbl_name_hl_chisq_out.;
   SET HL_ChiSq_dp;
		df=N_groups-2;
		ProbChiSq=1-PROBCHI( ChiSq, df);
		ProbChiSq_t=1-PROBCHI( ChiSq_t, df);
  RUN;

  PROC SORT DATA=&tbl_name_hl_chisq_out.;
   BY dp_cd;
   RUN;

  %MEND;
 