 /*HEADER------------------------------------------------------------------------
|
| PROGRAM:  calc_matrix_product.sas (macro)
||
| PURPOSE:  This macro computes matrix product of two matrices.
|
|
| MACRO PARAMETERS:
|
	 INPUT:    
          &tbl_name_left_in   The name of the input left table for the matrix product. Must be specified
|	     &tbl_name_right_in  The name of the input right table for the matrix product. Must be specified.
										 
		 &tbl_name_out          The name of the output table with the matrix product. If not specified defaults
										    to the table name: MATRIX_PRODUCT. The columns' names of the
                                            output table are always the same as column names of the right table. 
                                            The table can also have an optional first column which identifies a row in
                                            the left table. (see below description of the optional parameter 
                                           &tbl_left_col_with_row_nm .In short

 &tbl_right_is_symmetric_flag  Optional. Can be used to improve performance when right matrix is symmetric.
											   Value equal 1 suppresses the step which trasposes right matrix. 
                                               The default is 0. 

 &tbl_left_col_with_row_nm	 Optional. Allows to specify an additional variable which identifies a row in the left table.
											  For example, in the the column matrix of regression coefficient the variable
											 can identify corresponding covariate. This variable does not affect any 
                                             calculations and is simply passed to the the output table .
											 It can be useful to identify rows by some meaningful name rather than by row number only.
  
&tbl_right_col_with_row_nm  Similar to tbl_left_col_with_row_nm but it is not past to the output. It tells 
                                                the macro which column is not part of the matrix and has to be excluded 
                                                 from the right table before calculation start.

   	USAGE EXAMPLES:
 
%calc_matrix_product(tbl_name_left_in=A, tbl_name_right_in =B,tbl_name_out=AB_Y1);

%calc_matrix_product(tbl_name_left_in=A1, tbl_name_right_in =B,tbl_name_out=AB_Y2
						                            ,tbl_left_col_with_row_nm=ROW_NM);

%calc_matrix_product(tbl_name_left_in=A1, tbl_name_right_in =B_symm,tbl_name_out=AA_Y3
						                            ,tbl_left_col_with_row_nm=ROW_NM,tbl_right_is_symmetric_flag=1);
 
	First version:  02/02/2016, Yury Vilk. 
+------------------------------------------------------------------------HEADER*/

%MACRO calc_matrix_product(tbl_name_left_in=
                                                   ,tbl_name_right_in =
                                                    ,tbl_name_out=MATRIX_PRODUCT
						                            ,tbl_left_col_with_row_nm=
                                                    ,tbl_right_col_with_row_nm=
                                                    ,tbl_left_is_symmetric_flag=0
                                                    ,tbl_right_is_symmetric_flag=0);
   %GLOBAL DEBUG_FLAG2;
				   /* Code to suppress unnessary messages in the log except when in the debug mode.*/
 %LOCAL _DEBUG_OPTIONS_L;
 %LET _DEBUG_OPTIONS_L=%sysfunc(getoption(symbolgen)) %sysfunc(getoption(mprint)) %sysfunc(getoption(mlogic)) %sysfunc(getoption(notes)) %sysfunc(getoption(source2));

 %IF &DEBUG_FLAG2. NE  Y %THEN %DO;OPTIONS NOMLOGIC NOMPRINT NOSYMBOLGEN NONOTES NOSOURCE2; %END;

   %PUT;
   %PUT -----   Start of the macro calc_matrix_product. ;
   %PUT;

 
%LOCAL  tbl_left_nrows tbl_left_ncols	tbl_left_col_list tbl_left_rename_str tbl_left_col_with_row_nm_len
				  tbl_right_nrows tbl_right_ncols	tbl_right_col_list tbl_right_rename_str tbl_right_col_with_row_nm_len
				  _tbl_left_in _tbl_right_in pos_l pos_r _drop_col1 tbl_name_left_in_root tbl_name_right_in_root
			;	

	  %LET pos_l=%INDEX(%STR(&tbl_name_left_in.),%STR(%());
	  %LET pos_r=%INDEX(%STR(&tbl_name_right_in.),%STR(%());

 
%IF &pos_l. >1 %THEN %LET tbl_name_left_in_root=%SUBSTR(&tbl_name_left_in.,1,%EVAL(&pos_l.-1) );
%ELSE  							  %LET tbl_name_left_in_root=&tbl_name_left_in.;

%IF &pos_r. >1 %THEN %LET tbl_name_right_in_root=%SUBSTR(&tbl_name_right_in.,1,%EVAL(&pos_r.-1) );
%ELSE  							  %LET tbl_name_right_in_root=&tbl_name_right_in.;

%IF      %LENGTH(&tbl_name_left_in.)=0  OR %sysfunc(exist(&tbl_name_left_in_root.))=0
    OR %LENGTH(&tbl_name_right_in.)=0 OR  %sysfunc(exist(&tbl_name_right_in_root.))=0
 %THEN 
			%DO; 
   %PUT *************** ERROR NOT ALL REQUIERED INPUT PARAMETERS ARE SPECIFIED OR EXIST ***************;
   %PUT tbl_name_left_in=&tbl_name_left_in.  tbl_name_right_in=&tbl_name_right_in.  tbl_name_out=&tbl_name_out.; 
   %PUT *************** MACRO  ENDED UNSUCCESSFULLY *********;
  %GOTO endmac; 
   %END;

			/* If there are already tables listed below then delete them. They will be created by this macro. */
					PROC DATASETS NOLIST NOWARN;
					  DELETE _cont_left _cont_right _tmp _tr_tbl_right;
					RUN; 
                   QUIT;

		%LET _drop_col1=;
		
	 %IF       %LENGTH(&tbl_right_col_with_row_nm.) >0 
     %THEN  %LET _drop_col1= %STR(DROP &tbl_right_col_with_row_nm.);;

	 %IF &pos_r. >0 OR %LENGTH(&tbl_right_col_with_row_nm.) >0 %THEN
              						%DO;
			DATA _tbl_right_in;
			   SET &tbl_name_right_in.;
			   &_drop_col1 ;
			RUN;
		                  %LET _tbl_right_in=_tbl_right_in;
		       						%END;
		%ELSE 		 %LET _tbl_right_in=&tbl_name_right_in.;

		%IF &pos_l. >0  %THEN
              								%DO;
			DATA _tbl_left_in;
			   SET &tbl_name_left_in.;
			RUN;
		                  %LET _tbl_left_in=_tbl_left_in;
		       								%END;
		%ELSE 		 %LET _tbl_left_in=&tbl_name_left_in.;

	PROC CONTENTS DATA=&_tbl_left_in.	 OUT=_cont_left (KEEP=NOBS DELOBS NAME VARNUM LENGTH TYPE )	
		NOPRINT; 
	RUN;

	PROC CONTENTS DATA=&_tbl_right_in.	OUT=_cont_right(KEEP=NOBS DELOBS NAME VARNUM LENGTH TYPE )  
		NOPRINT; 
   RUN;

   		 PROC  SQL NOPRINT;
			SELECT MAX(CASE 
								 WHEN     TYPE=1 
									THEN  CATX(" ", "LENGTH", NAME, LENGTH)
							     WHEN    TYPE=2 
									THEN  CATX(" ", "LENGTH", NAME, "$", LENGTH)
	                          	  ELSE 	" "
						   END)  
			 INTO   : tbl_left_col_with_row_nm_len
				FROM  _cont_left
				 WHERE UPCASE(NAME)= UPCASE("&tbl_left_col_with_row_nm." )
				 ;
		QUIT;

	/* Ignore tbl_left_col_with_row_nm if it is specified in the macro cal but does not exist in the left table. */
%IF %LENGTH(&tbl_left_col_with_row_nm_len.)=0 %THEN %LET tbl_left_col_with_row_nm= ;

	PROC  SQL NOPRINT;
	  	SELECT 	STRIP(PUT(MAX(NOBS) -MAX(DELOBS), 16.))
						,STRIP(PUT(COUNT(*),6.))
			 INTO  :tbl_left_nrows
					  , :tbl_left_ncols
				FROM  _cont_left
				WHERE UPCASE(NAME) NE UPCASE("&tbl_left_col_with_row_nm.") 
				     AND TYPE=1
				 ;

			SELECT STRIP(NAME)
 	 				   ,CATS(NAME,"=","COL",VARNUM)
					   ,CATS("COL",VARNUM)
			 INTO  :tbl_left_col_list SEPARATED BY " " 
					   ,:tbl_left_rename_str SEPARATED BY " " 
					   ,:tbl_left_col_list_after_rename SEPARATED BY " "
				FROM  _cont_left
				  WHERE UPCASE(NAME) NE UPCASE("&tbl_left_col_with_row_nm.")
				        AND TYPE=1
				 ORDER BY VARNUM
				 ; 
	 QUIT;

	 PROC  SQL NOPRINT;
	  	SELECT 	STRIP(PUT(MAX(NOBS) -MAX(DELOBS), 16.))
						,STRIP(PUT(MAX(VARNUM),6.))
			 INTO  :tbl_right_nrows
					  , :tbl_right_ncols
				
				FROM  _cont_right
				 ;
			SELECT STRIP(NAME)
 						  ,CATS(NAME,"=_","COL",VARNUM)
						  ,CATS("_PROD",VARNUM,"=", NAME)
			 INTO  :tbl_right_col_list SEPARATED BY " " 
			 			,:tbl_right_rename_str SEPARATED BY " "
						,:tbl_right_rename_back_str SEPARATED BY " " 
				FROM  _cont_right
				ORDER BY VARNUM
				 ;
	 QUIT;

%IF &tbl_left_ncols. NE &tbl_right_nrows. %THEN 
						 %DO;
	%PUT  The matrises are incompatible for multiplication tbl_left_rows NE  tbl_right_cols. Exiting macro; 
	%PUT  tbl_left_nrows=&tbl_left_nrows.  tbl_right_ncols=&tbl_right_ncols.;
					%GOTO endmac;
						 %END;

 			  /* Don't do transpose left matrix if it is known to be symmetric.*/
 %IF &tbl_right_is_symmetric_flag. =0 %THEN  
							%DO;
				%LET _tr_tbl_right=_tr_tbl_right;

PROC TRANSPOSE DATA=&_tbl_right_in.
								   OUT=_tr_tbl_right(DROP=_NAME_) 
							        PREFIX=_COL;
RUN;
							%END;
%ELSE  %DO; %LET _tr_tbl_right=&_tbl_right_in.(RENAME=(&tbl_right_rename_str.));	%END;


	/*Making Cartesian product of the left and transposed right matrices.
	  The advantage of using data step method for this is that it preserve rows order. */
data _tmp;
    do until(_eof);
        set &_tbl_left_in.(RENAME=(&tbl_left_rename_str.))
				end=_eof;
        do i=1 to _nobs;
            set &_tr_tbl_right nobs=_nobs point=i;
            output;
        end;
    end;
run;

data &tbl_name_out(RENAME=(&tbl_right_rename_back_str.));
   &tbl_left_col_with_row_nm_len. ;
    /*new columns for products, equal to number of columns in the right matrix*/
    array _prod{&tbl_right_ncols.};

	ARRAY l_cols{&tbl_left_ncols.}           &tbl_left_col_list_after_rename.;
	ARRAY tr_r_cols{&tbl_right_nrows.} _COL1- _COL&tbl_right_nrows.;

    do j=1 to &tbl_right_ncols.;
        _prod{j]=0;
        set _tmp;

		        /*multiply corresponding pairs of columns*/
        do i=1 to &&tbl_left_ncols.;
            _prod{j]+l_cols{i]*tr_r_cols{i};
        end;
    end;
    output;
     keep &tbl_left_col_with_row_nm. _prod:;
run;

%IF &SYSERR.=0 %THEN %DO; %PUT Matrix product was calculated succesfuly.;  %END;

  %endmac:
 
			/* Delete temp tables created by this macro except when in the debug mode. */
	  %IF &DEBUG_FLAG2. NE  Y %THEN 
							%DO;
					PROC DATASETS NOLIST NOWARN;
					  DELETE _cont_left _cont_right _tmp _tr_tbl_right;
					RUN; 
                   QUIT;
						   %END;
		%ELSE
						  %DO;
	   %PUT tbl_left_nrows=&tbl_left_nrows. tbl_left_ncols=&tbl_left_ncols. ; 
		%PUT tbl_left_col_list=&tbl_left_col_list. tbl_left_rename_str=&tbl_left_rename_str	;
		%PUT tbl_left_col_with_row_nm_len=&tbl_left_col_with_row_nm_len	;
		%PUT tbl_left_col_list_after_rename=&tbl_left_col_list_after_rename. ;

	  %PUT tbl_right_nrows=&tbl_right_nrows.  tbl_right_ncols=&tbl_right_ncols. ;
	  %PUT tbl_right_col_list=&tbl_right_col_list. tbl_right_rename_str=&tbl_right_rename_str.	;
	  %PUT tbl_right_rename_back_str=&tbl_right_rename_back_str.;

		%PUT  _tbl_right_in=&_tbl_right_in.  _tbl_left_in=&_tbl_left_in.  ; ;
						  %END;

	%PUT;
   %PUT -----   End of the macro calc_matrix_product. ;
   %PUT;

	OPTIONS &_DEBUG_OPTIONS_L.;

%MEND; /*calc_matrix_product*/

